from recipe import SETTINGS
from recipe.dynamic_extensions import DynamicExtensionBase
from recipe_caching.mappers import FromCache


class CachingQueryHook(DynamicExtensionBase):
    """A stevedore based DynamicExtension that sets the proper options for a
    cached query
    """
    def __init__(self, recipe_parts):
        """Initializes a new CachingQueryHook
        :param recipe_parts: the components of a recipe
        :type key: dict
        """
        super(CachingQueryHook, self).__init__(recipe_parts)

    def execute(self):
        """Adds the FromCache option to the query

        :return: all the recipe components in recipe_parts format
        :rtype: dict
        """
        cache_target = getattr(SETTINGS, 'CACHE_REGION', 'default')
        cache_prefix = getattr(SETTINGS, 'CACHE_PREFIX', None)
        self.recipe_parts['query'] = self.recipe_parts['query'].options(
            FromCache(cache_target, cache_prefix=cache_prefix))
        return self.recipe_parts



""" mappers.py
Introduces a query option called FromCache.

The three new concepts introduced here are:

 * FromCache - a query option that establishes caching
   parameters on a Query
 * RelationshipCache - a variant of FromCache which is specific
   to a query invoked during a lazy load.
 * _params_from_query - extracts value parameters from
   a Query.
"""
from sqlalchemy.orm.interfaces import MapperOption


class FromCache(MapperOption):
    """Specifies that a Query should load results from a cache."""

    propagate_to_loaders = False

    def __init__(self, region="default", cache_key=None, cache_prefix=None):
        """Construct a new FromCache.
        :param region: the cache region.  Should be a
        region configured in the dictionary of dogpile
        regions.
        :param cache_key: optional.  A string cache key
        that will serve as the key to the query.   Use this
        if your query has a huge amount of parameters (such
        as when using in_()) which correspond more simply to
        some other identifier.
        """
        self.region = region
        self.cache_key = cache_key
        self.cache_prefix = cache_prefix

    def process_query(self, query):
        """Process a Query during normal loading operation."""
        query._cache_region = self


class RelationshipCache(MapperOption):
    """Specifies that a Query as called within a "lazy load"
       should load results from a cache."""
    propagate_to_loaders = True

    def __init__(self, attribute, region="default", cache_key=None):
        """Construct a new RelationshipCache.

        :param attribute: A Class.attribute which
        indicates a particular class relationship() whose
        lazy loader should be pulled from the cache.

        :param region: name of the cache region.

        :param cache_key: optional.  A string cache key
        that will serve as the key to the query, bypassing
        the usual means of forming a key from the Query itself.

        """
        self.region = region
        self.cache_key = cache_key
        self._relationship_options = {
            (attribute.property.parent.class_, attribute.property.key): self
        }

    def process_query_conditionally(self, query):
        """Process a Query that is used within a lazy loader.

        (the process_query_conditionally() method is a SQLAlchemy
        hook invoked only within lazyload.)

        """
        if query._current_path:
            mapper, prop = query._current_path[-2:]
            key = prop.key

            for cls in mapper.class_.__mro__:
                if (cls, key) in self._relationship_options:
                    relationship_option = self._relationship_options[
                        (cls, key)]
                    query._cache_region = relationship_option
                    break

    def and_(self, option):
        """Chain another RelationshipCache option to this one.

        While many RelationshipCache objects can be specified on a single
        Query separately, chaining them together allows for a more efficient
        lookup during load.

        """
        self._relationship_options.update(option._relationship_options)
        return self



import threading
import unicodedata

from dogpile.cache import make_region
from dogpile.cache.util import sha1_mangle_key


def async_creation_runner(cache, somekey, creator, mutex):
    """ Used by build_region as the async_creation_runner to create keys
    asyncronously. So dogpile.core:Lock when appropriate can avoid blocking.

    :param cache: the cache region we are writing too
    :type key: str
    :param key: the cache key
    :type key: str
    :param key: the creator function
    :type key: obj
    :param key: the mutex for our Lock
    :type key: obj

    :return: Nothing
    """
    def runner():
        try:
            value = creator()
            cache.set(somekey, value)
        finally:
            mutex.release()

    thread = threading.Thread(target=runner)
    thread.start()


def clean_unicode(value):
    """ Attempts to normalize a string to a nonunicode value

    :param value: the value to be cleaned of unicode characters
    :type value: str

    :return: a mangled key
    """
    try:
        cleaned_value = str(value)
    except UnicodeEncodeError:
        cleaned_value = unicodedata.normalize('NFKD', value).encode(
            'ascii', 'ignore')
        if not cleaned_value:
            raise ValueError("Could not find useful chars in the string")
    return cleaned_value


def unicode_sha1_mangle_key(key):
    """Used by mangle_key to sha1 mangle and clean_unicode from a supplied key

    :param key: the value to be mangled and prefixed, typically a SQL query
    :type key: str

    :return: a mangled key
    """
    return sha1_mangle_key(clean_unicode(key))


def mangle_key(key):
    """Used by build_region as the key_mangler for dogpile.cache. It prefixes,
    sha1 mangles, and cleans unicode from a supplied key

    :param key: the value to be mangled and prefixed, typically a SQL query
    :type key: str

    :return: a prefixed and mangled key
    """
    base = 'recipe_cache:'
    try:
        prefix, key = key.split(':', 1)
    except ValueError:
        prefix = None

    if prefix:
        base += '{}'.format(prefix)
    else:
        raise ValueError(key)

    return '{}:{}'.format(base, unicode_sha1_mangle_key(key))


def build_region(region_type='redis', region_args={}):
    """An implementation of dogpile.caches make_region that provides a thread
    based async_creation_runner and a prefix sha1 key_mangler.

    :param region_type: a string of the caching backend to use
    :type region_type: str
    :param region_args: a dictionary of configuration values for the backend
    :type region_type: dict

    :return: a cache region object, typically to be stored in recipe settings.
    """
    return make_region(async_creation_runner=async_creation_runner,
                key_mangler=mangle_key).configure(
        'dogpile.cache.' + region_type,
        arguments=region_args

    )



from recipe.oven import get_oven


def test_create_caching_oven():
    oven = get_oven('sqlite://', name='caching')
    assert oven.engine.driver == 'pysqlite'
    assert oven.Session.kw['bind'] == oven.engine


def test_create_caching_oven_no_target():
    oven = get_oven(name='caching')
    assert oven.engine is None
    assert oven.Session is None
